/*******************************************************************************
    Copyright (c) 2011 Yahoo! Inc. All rights reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License. See accompanying LICENSE file.

    The Initial Developer of the Original Code is Shravan Narayanamurthy.
******************************************************************************/
/*
 * MedLDA_Model_Tester.cpp
 *
 *  Created on: 06-Jan-2011
 *      
 */
#include <mpi.h>
#include <algorithm>

#include "MedLDA_Model_Tester.h"
#include "document.pb.h"
#include "sampler.h"
#include "Context.h"
#include "glog/logging.h"
#include "TopicLearner/Dirichlet.h"

namespace medlda {

MedLDA_Model_Tester::MedLDA_Model_Tester(TypeTopicCounts& ttc,
        Parameter& alpha, Parameter& beta, Parameter* eta_, 
        const size_t num_docs, bool no_init) :
        _ttc(ttc), _alpha(alpha), _beta(beta), _eta(eta_), _num_docs(num_docs) {
    if (no_init)
        return;
    Context& context = Context::get_instance();
    string input_w = context.get_string("input_w");
    string input_t = context.get_string("input_t");
    string output_t = context.get_string("output_t");
    set_up_io(input_w, input_t, output_t);
    _num_words = _ttc.get_num_words();
    _num_topics = _ttc.get_num_topics();

    string ttc_dumpfile = context.get_string("ttc_dumpfile");
    string eta_dumpfile = context.get_string("eta_dumpfile");

    int num_dumps = context.get_int("numdumps");
    // if (num_dumps == 1) {
    //     _ttc.initialize_from_dump(ttc_dumpfile);
    // } else {
    int num_words_per_client = _num_words / num_dumps;
    stringstream ss;
    for (int dump = 0; dump < num_dumps; dump++) {
      ss << ttc_dumpfile;
      ss << "." << dump;
      string dumpfile = ss.str();
      int offset = dump * num_words_per_client;
      _ttc.initialize_from_dump(dumpfile, NULL, NULL, offset);
      ss.str("");
    }
    // }

    double alpha_bar = context.get_double("alpha");
    //LOG(WARNING) << "Initializing Alpha vector from Alpha_bar = "
    //        << alpha_bar;
    _alpha.initialize_from_values(_num_topics, NULL, alpha_bar);
    //LOG(WARNING) << "Alpha vector initialized";

    double beta_flag = context.get_double("beta");
    //LOG(WARNING) << "Initializing Beta Parameter from specified Beta = "
    //        << beta_flag;
    double beta_value[1] = { beta_flag };
    _beta.initialize_from_values(1, beta_value, _num_words * beta_flag);
    //LOG(WARNING) << "Beta param initialized";

    //LOG(WARNING) << "Initializing Eta vector with zeros";
    _num_label = context.get_int("num_label");
    _eta->initialize_eta_from_dump(eta_dumpfile, _num_label);
    //LOG(WARNING) << "Eta vector initialized";

    //Initialize the boost RNG
    for (int i = 0; i < NUM_RNGS; i++) {
        generators[i] = new base_generator_type();
        uni_dists[i] = new uniform_real<> (0, 1);
        unif01[i] = new variate_generator<base_generator_type&,
                boost::uniform_real<> > (*(generators[i]), *(uni_dists[i]));
        //Throw away a few initial values
        for (int j = 0; j < 25; j++) {
            (*(unif01[i]))();
        }
    }
    rng_ind = 0;

    //For global values
    initial_c = context.get_double("initial_c");
    nusqinv = context.get_double("nusqinv");
    delta_ell = context.get_double("delta_ell");
    num_accurate = 0;

    phi = new double*[_num_topics];
    for (int t = 0; t < _num_topics; ++t) {
        phi[t] = new double[_num_words];
        memset(phi[t], 0, sizeof(double)*_num_words);
    }

    topicCounts current_topic_counts(_num_topics); 
    atomic<topic_t> *tokens_per_topic = new atomic<topic_t> [_num_topics]; 
    _ttc.get_counts(tokens_per_topic); 
    double *pphi;
    double b = _beta.values[0];
    for (int t = 0; t < _num_topics; ++t) {
        pphi = phi[t];
        double div = tokens_per_topic[t] + _beta.sum;
        for (int w = 0; w < _num_words; ++w) {
            _ttc.get_counts(w, &current_topic_counts);
            pphi[w] = (current_topic_counts.get_counts(t) + b) / div;
        }
    }

    gibbs_converged = context.get_double("gibbs_converged");
    gibbs_max_iter = context.get_int("gibbs_max_iter");
    sampling_lag = context.get_int("sampling_lag");

    pred_list.resize(_num_docs);

    ignore_old_topic = false;
}

MedLDA_Model_Tester::~MedLDA_Model_Tester() {
    for (int i = 0; i < NUM_RNGS; i++) {
        delete generators[i];
        delete uni_dists[i];
        delete unif01[i];
    }
    for (int t = 0; t < _num_topics; ++t) {
        delete[] phi[t];
    }
    delete[] phi;
    release_io();
}

void MedLDA_Model_Tester::set_up_io(string input_w, string input_t, string output_t) {
    _wdoc_rdr = new DocumentReader(input_w);
    _tdoc_rdr = new DocumentReader(input_t);

    _tdoc_writer = new DocumentWriter(output_t);
}

void MedLDA_Model_Tester::release_io() {
    if (_wdoc_rdr)
        delete _wdoc_rdr;
    if(_tdoc_rdr!=NULL) delete _tdoc_rdr;
    if (_tdoc_writer)
        delete _tdoc_writer;
}

google::protobuf::Message* MedLDA_Model_Tester::allocate_document_buffer(
        size_t num_docs) {
    return new LDA::MedLDA_document[num_docs];
}

void MedLDA_Model_Tester::deallocate_document_buffer(
        google::protobuf::Message* docs) {
    delete[] dynamic_cast<LDA::MedLDA_document*> (docs);
}

google::protobuf::Message* MedLDA_Model_Tester::get_nth_document(
        google::protobuf::Message* docs, size_t n) {
    return dynamic_cast<LDA::MedLDA_document*> (docs) + n;
}

void* MedLDA_Model_Tester::read(google::protobuf::Message& doc) {
    if (_wdoc_rdr->read(&doc) == -1)
        return NULL;
    LDA::MedLDA_document* wdoc = dynamic_cast<LDA::MedLDA_document*> (&doc);
    LDA::MedLDA_document tdoc;
    _tdoc_rdr->read(&tdoc);
    wdoc->MergeFrom(tdoc);
    MedLDA_update_t *upd = new MedLDA_update_t;
    upd->doc = wdoc;
    return upd;
}

void* MedLDA_Model_Tester::test(void* token) {
    MedLDA_update_t* upd = (MedLDA_update_t*) token;
    LDA::MedLDA_document* doc = upd->doc;
    int num_words_in_doc = doc->body_size();

    if (doc->docid() % 1000 == 0) 
        LOG(WARNING) << "doc " << doc->docid();

    int *nd = new int[_num_topics];
    memset(nd, 0, sizeof(int)*_num_topics);
    int **nw = new int*[_num_topics];
    for (int t = 0; t < _num_topics; ++t) {
        nw[t] = new int[_num_words];
        memset(nw[t], 0, sizeof(int)*_num_words);
    }

    for (int n = 0; n < num_words_in_doc; n++) {
        topic_t topic = doc->topic_assignment(n);
        word_t word = doc->body(n);
        ++nd[topic];
        ++nw[topic][word];
    }

    //RNG index
    int rng_index = (num_words_in_doc + _num_topics + rng_ind++) & RNG_MASK; 
    topicCounts current_topic_counts(_num_topics); 
    atomic<topic_t> *tokens_per_topic = new atomic<topic_t> [_num_topics]; 
    _ttc.get_counts(tokens_per_topic); 

    //Params
    double *alpha_array = _alpha.values;
    register double reg_beta = _beta.values[0];
    register double reg_beta_sum = _beta.sum;

    //Alloc probs here
    double *prob = new double[_num_topics];

    // burn in
    double iter = 0;
    double converged = 1;
    double old = 0;
    while (converged > gibbs_converged && iter < gibbs_max_iter) {
        for (int n = 0; n < num_words_in_doc; ++n) {
            register topic_t reg_old_topic = doc->topic_assignment(n);
            register word_t reg_word = doc->body(n);
    
            _ttc.get_counts(reg_word, &current_topic_counts);

            //Decrement
            --nw[reg_old_topic][reg_word];
            --nd[reg_old_topic];

            //Accumulate probs
            prob[0] = (current_topic_counts.get_counts(0) + nw[0][reg_word] + reg_beta) / 
                    (tokens_per_topic[0] + nd[0] + reg_beta_sum) *
                    (nd[0] + alpha_array[0]);
            for (topic_t t = 1; t < _num_topics; ++t) 
                prob[t] = prob[t-1] +
                        (current_topic_counts.get_counts(t) + nw[t][reg_word] + reg_beta) / 
                        (tokens_per_topic[t] + nd[t] + reg_beta_sum) *
                        (nd[t] + alpha_array[t]);

            //Sample
            register topic_t reg_new_topic = 0;
            double rd = (*(unif01[rng_index]))() * prob[_num_topics-1];
            for(; prob[reg_new_topic] < rd; ++reg_new_topic);

            //Increment
            ++nd[reg_new_topic];
            ++nw[reg_new_topic][reg_word];

            //Note the new topic assignment in doc
            doc->set_topic_assignment(n, reg_new_topic);

            LOG_IF(FATAL,reg_new_topic == -1 || reg_new_topic >= _num_topics)<<"Error in Tester. Newtopic generated was: " << reg_new_topic << ". Quitting";
        } // end of iterating over N words

        //compute lhood
        double div = _alpha.sum + num_words_in_doc;
        double dlhood = 0, dlword;
        for (int n = 0; n < num_words_in_doc; ++n) {
            dlword = 0;
            word_t word = doc->body(n);
            for (int t = 0; t < _num_topics; ++t) {
                dlword += phi[t][word] * (nd[t] + alpha_array[t]) / div;
            }
            dlhood -= log(dlword);
        }
        converged = fabs((old - dlhood) / old);
        old = dlhood;
        ++iter;
    } // end of while converge

    //Sampling suff topics
    int *suff_topics = new int[_num_topics];
    memset(suff_topics, 0, sizeof(int)*_num_topics);
    for (int i = 0; i < sampling_lag; ++i) {
        for (int n = 0; n < num_words_in_doc; ++n) {
            register topic_t reg_old_topic = doc->topic_assignment(n);
            register word_t reg_word = doc->body(n);
            _ttc.get_counts(reg_word, &current_topic_counts);

            //Decrement
            --nw[reg_old_topic][reg_word];
            --nd[reg_old_topic];

            //Accumulate probs
            prob[0] = (current_topic_counts.get_counts(0) + nw[0][reg_word] + reg_beta) / 
                    (tokens_per_topic[0] + nd[0] + reg_beta_sum) *
                    (nd[0] + alpha_array[0]);
            for (topic_t t = 1; t < _num_topics; ++t) 
                prob[t] = prob[t-1] +
                        (current_topic_counts.get_counts(t) + nw[t][reg_word] + reg_beta) / 
                        (tokens_per_topic[t] + nd[t] + reg_beta_sum) *
                        (nd[t] + alpha_array[t]);

            //Sample
            register topic_t reg_new_topic = 0;
            double rd = (*(unif01[rng_index]))() * prob[_num_topics-1];
            for(; prob[reg_new_topic] < rd; ++reg_new_topic);

            //Increment
            ++nd[reg_new_topic];
            ++nw[reg_new_topic][reg_word];

            //Note the new topic assignment in doc
            doc->set_topic_assignment(n, reg_new_topic);
            ++suff_topics[reg_new_topic];

            LOG_IF(FATAL,reg_new_topic == -1 || reg_new_topic >= _num_topics)<<"Error in Tester. Newtopic generated was: " << reg_new_topic << ". Quitting";
        } // end of iter over n words
    } // end of sampling lag

    //Do prediction
    double score;
    double *eta_array;
    int pred = -1;
    for (int y = 0; y < _num_label; ++y) { // TODO: multi label
        score = 0;
        eta_array = _eta[y].values;
        for (int t = 0; t < _num_topics; ++t) {
            score += eta_array[t] 
                    * ((double)suff_topics[t]/sampling_lag + alpha_array[t])
                    / (num_words_in_doc + _alpha.sum);
        }
        label_score_t ls;
        ls.label = y;
        ls.score = score;
        {
            tbb::mutex::scoped_lock lock(mut);
            pred_list[doc->docid()].push_back(ls);
        }
    }

    //Free resources
    delete[] tokens_per_topic;
    delete[] suff_topics;
    delete[] nd;
    for (int t = 0; t < _num_topics; ++t) {
        delete[] nw[t];
    }
    delete[] nw;
    delete[] prob;

    return upd;
}

void* MedLDA_Model_Tester::sort(void*) {
    LOG(FATAL) << "Sample method called for tester";
}

void* MedLDA_Model_Tester::sample(void*) {
    LOG(FATAL) << "Sample method called for tester";
}

void* MedLDA_Model_Tester::sample_eta(void*) {
    LOG(FATAL) << "Sample method called for tester";
}

void* MedLDA_Model_Tester::update(void* token) {
    LOG(FATAL) << "Sample method called for tester";
}

void* MedLDA_Model_Tester::eval(void* token, double& ret) {
    LOG(FATAL) << "Sample method called for tester";
}

void* MedLDA_Model_Tester::predict(void* token) {
    LOG(FATAL) << "Sample method called for tester";
}

void* MedLDA_Model_Tester::optimize(void*) {
    LOG(FATAL) << "Optimize method called for tester";
}

void MedLDA_Model_Tester::write(void* token) {
    MedLDA_update_t* upd = (MedLDA_update_t*) token;
    LDA::MedLDA_document& doc = *(upd->doc);
    doc.clear_body();
    doc.clear_gnd_label();
    LOG_IF(FATAL, !_tdoc_writer->write(doc) )<< "Couldn't write to the stream. Quitting..." << endl;
    delete upd;
}

bool compare_label_score(const label_score_t &l, const label_score_t &r) {
  return l.score > r.score;
}

void MedLDA_Model_Tester::iteration_done() {
    Context& context = Context::get_instance();
    string input_w = context.get_string("input_w");
    string input_t = context.get_string("input_t");
    string output_t = context.get_string("output_t");

    //string cmd = "cp " + output_t + " " + input_t;
    string cmd = "mv " + output_t + " " + input_t;
    system(cmd.c_str());

    release_io();
    set_up_io(input_w, input_t, output_t);

    int world_rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    stringstream ss;
    ss << "." << world_rank;
    string dotrank = ss.str();

    //save pred list
    string predfile = context.get_string("predfile") + dotrank;
    FILE *fp = fopen(predfile.c_str(), "w");
    for (int d = 0; d < _num_docs; ++d) {
        std::sort(pred_list[d].begin(), pred_list[d].end(), compare_label_score);
        for (int i = 0; i < pred_list[d].size() && i < 10; ++i) {
            if (i == 0) {
                fprintf(fp, "%d %g", pred_list[d][i].label, pred_list[d][i].score);
            } else {
                fprintf(fp, " %d %g", pred_list[d][i].label, pred_list[d][i].score);
            }
        }
        if (pred_list[d].size() == 0UL) {
            fprintf(fp, "-1");
        }
        fprintf(fp, "\n");
    }
    fclose(fp);
}

void MedLDA_Model_Tester::print_result() {
}

}
