/*******************************************************************************
    Copyright (c) 2011 Yahoo! Inc. All rights reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License. See accompanying LICENSE file.

    The Initial Developer of the Original Code is Shravan Narayanamurthy.
******************************************************************************/
/*
 * MedLDA_Model_Trainer.cpp
 *
 *  Created on: 06-Jan-2011
 *      
 */
#include <mpi.h>

#include "MedLDA_Model_Trainer.h"
#include "document.pb.h"
#include "sampler.h"
#include "Context.h"
#include "glog/logging.h"
#include "TopicLearner/Dirichlet.h"

#include "utils.h"
#include <math.h>
#include <unistd.h>

namespace medlda {
//#define SSE_OPT
#ifdef SSE_OPT
#include <nmmintrin.h>


typedef union {
  __m128 m;
  float f[4];
} m128_union;

#endif

MedLDA_Model_Trainer::MedLDA_Model_Trainer(TypeTopicCounts& ttc,
        Parameter& alpha, Parameter& beta, Parameter* eta_, const size_t num_docs) :
        _ttc(ttc), _alpha(alpha), _beta(beta), _eta(eta_), _num_docs(num_docs) { 
    Context& context = Context::get_instance();
    ignore_old_topic = context.get_bool("ignore_old_topic");
    string input_w = context.get_string("input_w");
    string input_t = context.get_string("input_t");
    string output_t = context.get_string("output_t");
    set_up_io(input_w, input_t, output_t);

    _num_words = _ttc.get_num_words();
    _num_topics = _ttc.get_num_topics();
    //LOG(WARNING) << "Initializing Word-Topic counts table from docs "
    //        << input_w << ", " << input_t << " using " << _num_words
    //        << " words & " << _num_topics << " topics.";
    _ttc.initialize_from_docs(input_w, input_t);
    //LOG(WARNING) << "Initialized Word-Topic counts table";

    double alpha_bar = context.get_double("alpha");
    //LOG(WARNING) << "Initializing Alpha vector from Alpha_bar = "
    //        << alpha_bar;
    _alpha.initialize_from_values(_num_topics, NULL, alpha_bar);
    //LOG(WARNING) << "Alpha vector initialized";

    double beta_flag = context.get_double("beta");
    //LOG(WARNING) << "Initializing Beta Parameter from specified Beta = "
    //        << beta_flag;
    double beta_value[1] = { beta_flag };
    _beta.initialize_from_values(1, beta_value, _num_words * beta_flag);
    //LOG(WARNING) << "Beta param initialized";

    //LOG(WARNING) << "Initializing Eta vector with zeros";
    _num_label = context.get_int("num_label");
    for (int y = 0; y < _num_label; ++y) {
        _eta[y].initialize_from_values(_num_topics, NULL, 0);
    }
    //LOG(WARNING) << "Eta vector initialized";

    //Initialize the boost RNG
    for (int i = 0; i < NUM_RNGS; i++) {
        generators[i] = new base_generator_type();
        uni_dists[i] = new uniform_real<> (0, 1);
        unif01[i] = new variate_generator<base_generator_type&,
                boost::uniform_real<> > (*(generators[i]), *(uni_dists[i]));
        //Throw away a few initial values
        for (int j = 0; j < 25; j++) {
            (*(unif01[i]))();
        }
    }
    rng_ind = 0;

    //For optimizer
    tau = 100;
    part_grads = new double [_num_topics];
    memset(part_grads,0,_num_topics*sizeof(double));
    part_grads_top_indep = -1 * tau * digamma(alpha.sum);

    //For global values
    initial_c = context.get_double("initial_c");
    nusqinv = context.get_double("nusqinv");
    delta_ell = context.get_double("delta_ell");
    MVGsampler = new MVGaussian();
    IGsampler = new InverseGaussian*[_num_docs];
    for (size_t i = 0; i < _num_docs; ++i) {
        IGsampler[i] = new InverseGaussian(1, 1);
    }
    num_accurate = 0;

    real_eta = new double[_num_label*_num_topics];
    real_kz = new double[_num_label*_num_topics];
    real_covinv = new double[_num_label*_num_topics*_num_topics];
    kappa_z = new double*[_num_label];
    covinv = new double**[_num_label];
    for (int y = 0; y < _num_label; ++y) {
        covinv[y] = new double*[_num_topics];
    }

    MPI_Comm_rank(MPI_COMM_WORLD, &taskid);
    MPI_Get_processor_name(processor_name, &name_len);
    if (taskid == 0) {
        recv_kz = new double[_num_label*_num_topics];
        recv_covinv = new double[_num_label*_num_topics*_num_topics];
        mu = new double[_num_topics];
        cov = new double*[_num_topics];
        lower_triangle = new double*[_num_topics];
        for (topic_t i = 0; i < _num_topics; ++i) {
            cov[i] = new double[_num_topics];
            lower_triangle[i] = new double[_num_topics];
        }
    }
    init_eta_components();
}

MedLDA_Model_Trainer::~MedLDA_Model_Trainer() {
    for (int i = 0; i < NUM_RNGS; i++) {
        delete generators[i];
        delete uni_dists[i];
        delete unif01[i];
    }

    release_io();

    for (size_t i = 0; i < _num_docs; ++i) {
        delete IGsampler[i];
    }
    delete[] IGsampler;
    delete MVGsampler;

    for (int y = 0; y < _num_label; ++y) {
        delete[] covinv[y];
    }
    delete[] kappa_z;
    delete[] covinv;
    delete[] real_kz;
    delete[] real_eta;
    delete[] real_covinv;
    if (taskid == 0) {
        delete[] recv_kz;
        delete[] recv_covinv;
        for (int i = 0; i < _num_topics; ++i) {
            delete[] cov[i];
            delete[] lower_triangle[i];
        }
        delete[] cov;
        delete[] lower_triangle;
        delete[] mu;
    }
}

void MedLDA_Model_Trainer::set_up_io(string input_w, string input_t,
        string output_t) {
    _wdoc_rdr = new DocumentReader(input_w);

    _tdoc_rdr = new DocumentReader(input_t);

    _tdoc_writer = new DocumentWriter(output_t);
}

void MedLDA_Model_Trainer::release_io() {
    delete _wdoc_rdr;
    if(_tdoc_rdr!=NULL) delete _tdoc_rdr;
    delete _tdoc_writer;
}

google::protobuf::Message* MedLDA_Model_Trainer::allocate_document_buffer(
        size_t num_docs) {
    return new LDA::MedLDA_document[num_docs];
}

void MedLDA_Model_Trainer::deallocate_document_buffer(
        google::protobuf::Message* docs) {
    delete[] dynamic_cast<LDA::MedLDA_document*> (docs);
}

google::protobuf::Message* MedLDA_Model_Trainer::get_nth_document(
        google::protobuf::Message* docs, size_t n) {
    return dynamic_cast<LDA::MedLDA_document*> (docs) + n;
}

void* MedLDA_Model_Trainer::read(google::protobuf::Message& doc) {
    if (_wdoc_rdr->read(&doc) == -1)
        return NULL;
    LDA::MedLDA_document* wdoc = dynamic_cast<LDA::MedLDA_document*> (&doc);
    LDA::MedLDA_document tdoc;
    _tdoc_rdr->read(&tdoc);
    wdoc->MergeFrom(tdoc);
    MedLDA_update_t *upd = new MedLDA_update_t;
    upd->doc = wdoc;
    return upd;
}

bool compare_z(const topic_n_t &l, const topic_n_t &r) {
    return l.z < r.z;
}

void* MedLDA_Model_Trainer::sort(void* token) {
    MedLDA_update_t* upd = (MedLDA_update_t*) (token);
    LDA::MedLDA_document* doc = upd->doc;
    int num_words_in_doc = doc->body_size();
    topic_n_t *tptr = new topic_n_t[num_words_in_doc];
    word_t *wptr = new word_t[num_words_in_doc];
    for (int n = 0; n < num_words_in_doc; ++n) {
        tptr[n].z = doc->topic_assignment(n);
        tptr[n].n = n;
        wptr[n] = doc->body(n);
    }
    std::stable_sort(tptr, tptr+num_words_in_doc, compare_z);
    for (int n = 0; n < num_words_in_doc; ++n) {
        doc->set_topic_assignment(n, tptr[n].z);
        doc->set_body(n, wptr[tptr[n].n]);
    }
    
    delete[] tptr;
    delete[] wptr;
    return upd;
}

void* MedLDA_Model_Trainer::sample(void* token) {
    MedLDA_update_t* upd = (MedLDA_update_t*) (token);
    vector<change_elem_t> *updates = 0;
    updates = new vector<change_elem_t> ();
    updates->reserve(EXP_NUM_WORDS_PER_DOC);
    sample_topics(upd, updates);

    return upd;
}

void MedLDA_Model_Trainer::sample_lambda(LDA::MedLDA_document *& doc,
    const int dlen, const topic_t *dptr, const int rng_index) {
  double dval, mean, cc;
  double *eta_array;
  cc = initial_c * initial_c;
  size_t idx = doc->docid();

  int gnd = -1;
  for (int y = 0; y < _num_label; ++y) {
    dval = 0;
    eta_array = _eta[y].values;
    for (topic_t t = 0; t < _num_topics; ++t) 
      dval += eta_array[t] * dptr[t];
    dval /= dlen;
    mean = initial_c / fabs(delta_ell - gnd*dval);
    IGsampler[idx]->setRNG(unif01[rng_index]);
    IGsampler[idx]->reset(mean, cc);
    doc->set_lambda(y, IGsampler[idx]->sample());
  }

  gnd = 1;
  int my_label;
  for (int i = 0; i < doc->gnd_label_size(); ++i) {
    dval = 0;
    my_label = doc->gnd_label(i);
    eta_array = _eta[my_label].values;
    for (topic_t t = 0; t < _num_topics; ++t) 
      dval += eta_array[t] * dptr[t];
    dval /= dlen;
    mean = initial_c / fabs(delta_ell - gnd*dval);
    IGsampler[idx]->setRNG(unif01[rng_index]);
    IGsampler[idx]->reset(mean, cc);
    doc->set_lambda(my_label, IGsampler[idx]->sample());
  }
}

bool MedLDA_Model_Trainer::compute_mrgterm_doc(const double gamma, 
        LDA::MedLDA_document *& doc, double *mrg) {
    memset(mrg, 0, _num_topics * sizeof(double));
    double dval, eta_t, lambda, gamma_lambda, gnd_ell_c, gnd_ell_c_pos, gnd_ell_c_neg;
    double half_gamma = 0.5 * gamma;
    double *eta_array;

    bool flag = false;
    for (int y = 0; y < _num_label; ++y) {
        eta_array = _eta[y].values;
        lambda = doc->lambda(y);
        gamma_lambda = gamma * lambda;
        gnd_ell_c = 0 - (delta_ell + initial_c/lambda); // y=-1
        for (topic_t t = 0; t < _num_topics; ++t) {
            eta_t = eta_array[t];
            mrg[t] += gamma_lambda * eta_t * (gnd_ell_c - half_gamma*eta_t); 
            if (mrg[t] > OVERFLOW_THREHOLD || mrg[t] < -OVERFLOW_THREHOLD) {
              flag = true;
            }
        }
    }
    int my_label;
    for (int i = 0; i < doc->gnd_label_size(); ++i) { 
        my_label = doc->gnd_label(i);
        eta_array = _eta[my_label].values;
        lambda = doc->lambda(my_label);
        gamma_lambda = gamma * lambda;
        gnd_ell_c_pos = (delta_ell + initial_c/lambda); // y=-1
        gnd_ell_c_neg = 0 - (delta_ell + initial_c/lambda); // y=-1
        for (topic_t t = 0; t < _num_topics; ++t) {
            eta_t = eta_array[t];
            mrg[t] -= gamma_lambda * eta_t * (gnd_ell_c_neg - half_gamma*eta_t); 
            mrg[t] += gamma_lambda * eta_t * (gnd_ell_c_pos - half_gamma*eta_t); 
            if (mrg[t] > OVERFLOW_THREHOLD || mrg[t] < -OVERFLOW_THREHOLD) {
              flag = true;
            }
        }
    }
    return flag;
}

void MedLDA_Model_Trainer::compute_delta(const Parameter *etas, 
    const double *disc, const double *dfactor, double *delta) {
  memset(delta, 0, sizeof(double)*_num_topics);
  double dval;
  double *eta_array;
  for (int y = 0; y < _num_label; ++y) {
    eta_array = etas[y].values;
    dval = disc[y] * dfactor[y];
    for (topic_t t = 0; t < _num_topics; ++t) {
      delta[t] += eta_array[t] * dval;
    }
  }
}

void MedLDA_Model_Trainer::sample_topics(MedLDA_update_t* upd, 
        std::vector<change_elem_t>* updates) {
    LDA::MedLDA_document* doc = upd->doc;
    int num_words_in_doc = doc->body_size();

    topic_t *document_topic_counts = new topic_t[_num_topics];
    memset(document_topic_counts, 0, sizeof(topic_t)*_num_topics);
    for (int k = 0; k < num_words_in_doc; k++) {
        topic_t topic = doc->topic_assignment(k);
        ++document_topic_counts[topic];
    }

    //RNG index
    int rng_index = (num_words_in_doc + _num_topics + rng_ind++) & RNG_MASK;

    //We can sample lambda here
    sample_lambda(doc, num_words_in_doc, document_topic_counts, rng_index);

    topicCounts current_topic_counts(_num_topics); //Storage for topic counts for the current word
    atomic<topic_t> *tokens_per_topic = new atomic<topic_t> [_num_topics]; //Storage for n(t)
    _ttc.get_counts(tokens_per_topic); //Get n(t)

    //Params
    double *alpha_array = _alpha.values;
    double *eta_array;
    register double reg_beta = _beta.values[0];
    register double reg_beta_sum = _beta.sum;

    //Prepare for mrg term
    double *mrg = new double[_num_topics];
    double *delta = new double[_num_topics];
    double gamma = 1.0 / (double)num_words_in_doc;
    bool log_space_flag = compute_mrgterm_doc(gamma, doc, mrg);

    double *dfactor = new double[_num_label];
    double *disc = new double[_num_label];
    double disc_temp;
    gamma = gamma * (gamma - 1);
    for (int y = 0; y < _num_label; ++y) {
        disc_temp= 0;
        eta_array = _eta[y].values;
        if (num_words_in_doc > 1) {
            for (topic_t t = 0; t < _num_topics; ++t)
                disc_temp += eta_array[t] * document_topic_counts[t];
            disc_temp /= (num_words_in_doc - 1);
        }
        disc[y] = disc_temp;
        dfactor[y] = gamma * doc->lambda(y);
    }

    //Alloc probs here
    double *prob = new double[_num_topics];
    int *word_topic_counts = new int[_num_topics];

    // gibbs sampling
    // log_space_flag = false; // NOTE(zhouli) shouldn't be always false
    topic_t prev = -1; 
    topic_t next = doc->topic_assignment(0);
    word_t prev_word = -1;
    bool changed = true;
    bool delta_changed = true;
    for (int n = 0; n < num_words_in_doc; ++n) {
        //For each word
        change_elem_t elem;
        elem.word = doc->body(n);
        elem.old_topic = next;
        next = (n == num_words_in_doc-1) ? -1 : doc->topic_assignment(n+1);
        register topic_t reg_old_topic = elem.old_topic;

        //Get the topic counts for current word
        _ttc.get_counts(elem.word, &current_topic_counts);

        //Delta with word n excluded
        if (reg_old_topic == prev) {
            //pass
        } else {
            if (num_words_in_doc > 1) {
                for (int y = 0; y < _num_label; ++y) {
                    eta_array = _eta[y].values;
                    disc[y] -= eta_array[reg_old_topic] / (num_words_in_doc - 1);
                }
            }
        }

        if (reg_old_topic == prev && !changed) {
            delta_changed = false;
        } else {
            compute_delta(_eta, disc, dfactor, delta);
            delta_changed = true;
        }

        //Decrement
        --document_topic_counts[reg_old_topic];
        --tokens_per_topic[reg_old_topic];

        if (!delta_changed && elem.word == prev_word) {
            //pass
        } else {
            // further decrement
            for (topic_t t = 0; t < _num_topics; ++t) {
                word_topic_counts[t] = current_topic_counts.get_counts(t);
                if (t == reg_old_topic && word_topic_counts[t] > 0)
                    --word_topic_counts[t];
            }

            if (log_space_flag) { // in log space
                for (topic_t t = 0; t < _num_topics; ++t) 
                    prob[t] = log((word_topic_counts[t] + reg_beta) / 
                            (tokens_per_topic[t] + reg_beta_sum) *
                            (document_topic_counts[t] + alpha_array[t])) +
                            (mrg[t] + delta[t]);
                double logsum = log_sum(prob, _num_topics);
                prob[0] = exp(prob[0] - logsum);
                for (topic_t t = 1; t < _num_topics; ++t) 
                    prob[t] = prob[t-1] + exp(prob[t] - logsum);
            } else { // original space
                prob[0] = (word_topic_counts[0] + reg_beta) / 
                        (tokens_per_topic[0] + reg_beta_sum) *
                        (document_topic_counts[0] + alpha_array[0]) *
                        exp(mrg[0] + delta[0]);
                for (topic_t t = 1; t < _num_topics; ++t) 
                    prob[t] = prob[t-1] +
                            (word_topic_counts[t] + reg_beta) / 
                            (tokens_per_topic[t] + reg_beta_sum) *
                            (document_topic_counts[t] + alpha_array[t]) *
                            exp(mrg[t] + delta[t]);
            }
        }
        //Sample
        register topic_t reg_new_topic = 0;
        double rd = (*(unif01[rng_index]))() * prob[_num_topics-1];
        for(; prob[reg_new_topic] < rd; ++reg_new_topic);

        //Increment
        ++document_topic_counts[reg_new_topic];
        ++tokens_per_topic[reg_new_topic];

        // adjust disc
        if (reg_new_topic == next) {
            //pass
        } else {
            if (num_words_in_doc > 1) {
                for (int y = 0; y < _num_label; ++y) {
                    eta_array = _eta[y].values;
                    disc[y] += eta_array[reg_new_topic] / (num_words_in_doc - 1);
                }
            }
        }

        doc->set_topic_assignment(n, reg_new_topic);
        elem.new_topic = reg_new_topic;
        LOG_IF(FATAL,elem.new_topic == -1 || elem.new_topic >= _num_topics)<<"Error in sampler. Newtopic generated was: " << elem.new_topic << ". Quitting";

        prev_word = elem.word;
        prev = reg_new_topic;
        changed = (reg_new_topic != reg_old_topic);

        if(updates!=NULL)
            updates->push_back(elem);
    } // end of iterating over N words

    //Free resources
    delete[] tokens_per_topic;
    delete[] document_topic_counts;
    delete[] mrg;
    delete[] delta;
    delete[] dfactor;
    delete[] disc;
    delete[] prob;
    delete[] word_topic_counts;

    upd->change_list = updates;
}

void* MedLDA_Model_Trainer::update(void* token) {
    MedLDA_update_t* upd = (MedLDA_update_t*) token;
    size_t sz = upd->change_list->size();
    for (size_t i = 0; i < sz; i++) {
        change_elem_t change = (*upd->change_list)[i];
        //		cout << "(" << dict->get_word(change.word) << "," << change.old_topic << "-" << change.new_topic << ")" << endl;
        //		cout << _ttc.print(change.word) << endl;
        _ttc.upd_count(change.word, change.old_topic, change.new_topic,
                ignore_old_topic);
        //		cout << _ttc.print(change.word) << endl;
    }
    if (sz != 0) {
        upd->change_list->clear();
    }
    if (upd->change_list != NULL)
        delete upd->change_list;

    return upd;
}

bool is_in_gnd(int y, LDA::MedLDA_document *doc) {
    for (int i = 0; i < doc->gnd_label_size(); ++i) {
        if (doc->gnd_label(i) == y) {
            return true;
        }
    }
    return false;
}

//debug
double summation(double *vec, int len) {
    double sum;
    for (int i = 0; i < len; ++i) {
        sum += vec[i];
    }
    return sum;
}

void* MedLDA_Model_Trainer::sample_eta(void* token) {
    double sampletime = 0;
    double samplestart = MPI_Wtime();

    MedLDA_update_t* upd = (MedLDA_update_t*) token;
    LDA::MedLDA_document* doc = upd->doc;
    int num_words_in_doc = doc->body_size();
    topic_t *document_topic_counts = new topic_t[_num_topics];
    memset(document_topic_counts, 0, sizeof(topic_t)*_num_topics);
    for (int k = 0; k < num_words_in_doc; k++) {
        topic_t topic = doc->topic_assignment(k);
        ++document_topic_counts[topic];
    }

    //Prepare for eta
    double kappa, lambda, temp1, temp2;
    double *ptr, *kz_ptr, **covinv_ptr;
    int binary, ndt;
    for (int y = 0; y < _num_label; ++y) {
        kz_ptr = kappa_z[y];
        covinv_ptr = covinv[y];
        binary = is_in_gnd(y, doc) ? 1 : -1;
        lambda = doc->lambda(y);
        kappa = binary * (delta_ell*lambda + initial_c) / num_words_in_doc;
        temp1 = lambda / (num_words_in_doc * num_words_in_doc);
        for (topic_t i = 0; i < _num_topics; ++i) {
            ndt = document_topic_counts[i];
            if (ndt > 0) {
                kz_ptr[i] += kappa * ndt;
                temp2 = temp1 * ndt;
                ptr = covinv_ptr[i];
                for (topic_t j = 0; j < _num_topics; ++j) {
                    ptr[j] += temp2 * document_topic_counts[j];
                }
            }
        }
        --num_unprepared_docs;
    }
    delete[] document_topic_counts;
    if (num_unprepared_docs != 0)
        return token;

    sampletime += MPI_Wtime() - samplestart;

    // MPI reduce
    //MPI_Barrier(MPI_COMM_WORLD);
    int err;
    double inittime, totaltime; 
    inittime = MPI_Wtime();
    err = MPI_Reduce(real_kz, recv_kz, _num_label*_num_topics, 
            MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    err = MPI_Reduce(real_covinv, recv_covinv, _num_label*_num_topics*_num_topics, 
            MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    totaltime = MPI_Wtime() - inittime;

    //debug
    //stringstream kzs;
    //for (int i = 0; i < 10; i++) {
    //    kzs << real_kz[i] << " ";
    //}
    //LOG(WARNING) << "send kz: " << kzs.str();
    //if (taskid == 0) {
    //    kzs.str("");
    //    for (int i = 0; i < 10; i++) {
    //        kzs << recv_kz[i] << " ";
    //    }
    //    LOG(WARNING) << "reduced kz: " << kzs.str();
    //}

    // if root
    double *eta_array;
    if (taskid == 0) {
        samplestart = MPI_Wtime();
        for (int y = 0; y < _num_label; ++y) {
            kappa_z[y] = recv_kz + y*_num_topics;
            for (topic_t i = 0; i < _num_topics; ++i) {
                covinv[y][i] = recv_covinv + y*_num_topics*_num_topics + i*_num_topics;
            }
        }
        for (int y = 0; y < _num_label; ++y) {
            kz_ptr = kappa_z[y];
            covinv_ptr = covinv[y];

            inverse_cholydec(covinv_ptr, cov, lower_triangle, _num_topics);
            for (topic_t t = 0; t < _num_topics; ++t)
                mu[t] = dotprod(cov[t], kz_ptr, _num_topics);

            eta_array = _eta[y].values;
            int rng_index = (_num_topics + rng_ind++) & RNG_MASK;
            MVGsampler->setRNG(unif01[rng_index]);
            MVGsampler->nextMVGaussianWithCholesky(mu, lower_triangle, eta_array, _num_topics);
            // _eta -> real
            memcpy(real_eta + y*_num_topics, eta_array, sizeof(double)*_num_topics);
        }
        sampletime += MPI_Wtime() - samplestart;
        LOG(WARNING) << "++++ SAMPLETIME: " << sampletime;
    }

    inittime = MPI_Wtime();
    err = MPI_Bcast(real_eta, _num_label*_num_topics, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    totaltime += MPI_Wtime() - inittime;
    if (taskid == 0) {
        LOG(WARNING) << "++++ MPITIME: " << totaltime;
    }

    // real -> _eta
    for (int y = 0; y < _num_label; ++y) {
        eta_array = _eta[y].values;
        memcpy(eta_array, real_eta + y*_num_topics, sizeof(double)*_num_topics);
    }

    init_eta_components();
    return upd;
}

//Init components for next usage 
void MedLDA_Model_Trainer::init_eta_components() 
{
    memset(real_eta, 0, _num_label*_num_topics*sizeof(double));
    memset(real_kz, 0, _num_label*_num_topics*sizeof(double));
    memset(real_covinv, 0, _num_label*_num_topics*_num_topics*sizeof(double));
    for (int y = 0; y < _num_label; ++y) {
        kappa_z[y] = real_kz + y*_num_topics;
        for (topic_t i = 0; i < _num_topics; ++i) {
            covinv[y][i] = real_covinv + y*_num_topics*_num_topics + i*_num_topics;
        }
    }
    for (topic_t i = 0; i < _num_topics; ++i) {
        for (int y = 0; y < _num_label; ++y) {
            covinv[y][i][i] = nusqinv;
        }
    }
    //for (int y = 0; y < _num_label; ++y) {
    //    memset(kappa_z[y], 0, _num_topics * sizeof(double));
    //}
    //for (topic_t i = 0; i < _num_topics; ++i) {
    //    for (int y = 0; y < _num_label; ++y) {
    //        memset(covinv[y][i], 0, _num_topics * sizeof(double));
    //        covinv[y][i][i] = nusqinv;
    //    }
    //}
    if (taskid == 0) {
        memset(recv_kz, 0, _num_label*_num_topics*sizeof(double));
        memset(recv_covinv, 0, _num_label*_num_topics*_num_topics*sizeof(double));
        memset(mu, 0, _num_topics * sizeof(double));
        for (topic_t i = 0; i < _num_topics; ++i) {
            memset(cov[i], 0, _num_topics * sizeof(double));
            memset(lower_triangle[i], 0, _num_topics * sizeof(double));
        }
    }
    num_unprepared_docs = _num_docs * _num_label;
}

void* MedLDA_Model_Trainer::predict(void* token) {
    MedLDA_update_t* upd = (MedLDA_update_t*) (token);
    LDA::MedLDA_document* doc = upd->doc;

    //Document statistics
    int num_words_in_doc = doc->body_size();
    topic_t *document_topic_counts = new topic_t[_num_topics];
    memset(document_topic_counts, 0, sizeof(topic_t)*_num_topics);
    for (int k = 0; k < num_words_in_doc; k++) {
        topic_t topic = doc->topic_assignment(k);
        ++document_topic_counts[topic];
    }

    double score, max_score = 0;
    double *alpha_array = _alpha.values;
    double *eta_array;
    int pred = -1;
    for (int y = 0; y < _num_label; ++y) {
        eta_array = _eta[y].values;
        score = 0;
        for (topic_t t = 0; t < _num_topics; ++t) 
            score += eta_array[t] * (double)(document_topic_counts[t] + alpha_array[t])
                    / (num_words_in_doc + _alpha.sum);
        if (pred == -1 || score > max_score) {
            pred = y;
            max_score = score;
        }
    }
    if (pred == doc->gnd_label(0))
        ++num_accurate;
    delete[] document_topic_counts;
    return upd;
}

void* MedLDA_Model_Trainer::optimize(void* token) {
    MedLDA_update_t* upd = (MedLDA_update_t*) token;
    LDA::MedLDA_document& doc = *(upd->doc);
    topic_t* local_topic_counts = new topic_t[_num_topics];
    topic_t* local_topic_index = new topic_t[_num_topics];
    int num_words_in_doc = doc.body_size();

    //Clear topic Counts
    memset(local_topic_counts, 0, _num_topics * sizeof(topic_t));
    memset(local_topic_index, 0, _num_topics * sizeof(topic_t));

    int non_zero_topics = 0;
    topic_t* ltcInd = local_topic_counts;
    topic_t* ltdInd = local_topic_index;
    if (!ignore_old_topic) {
        //Build the dense local_topic_counts with index
        for (int k = 0; k < num_words_in_doc; k++) {
            topic_t topic = doc.topic_assignment(k);
            ++local_topic_counts[topic];
        }

        for (topic_t t = 0; t < _num_topics; t++) {
            int n = *ltcInd;
            if (n != 0) {
                *ltdInd = t;
                ++ltdInd;
                ++non_zero_topics;
            }
            ++ltcInd;
        }
        //----- Local Topic Counts built
    }

    /************* Optimize **************/
    //		cout << "Optimizing: " << doc_index << endl;
    ++doc_index;
    if (doc_index != 0 && doc_index % tau == 0) {
        //Update global alphas with the partial
        //gradients that have been computed till now
        double* part_grads_index = part_grads;
        double *global_alphas_index = _alpha.values;
        double sum = 0.0;
        eta = pow((doc_index / tau) + 100, -0.5) * (0.1 / tau);
        for (topic_t t = 0; t < _num_topics; t++) {
            *global_alphas_index = std::max(
                    *global_alphas_index - eta * (*part_grads_index
                            + part_grads_top_indep), MIN_ALPHA);
            sum += *global_alphas_index;
            ++global_alphas_index;
            ++part_grads_index;
        }
        _alpha.sum = sum;

        //Reset the partial gradients
        memset(part_grads, 0, _num_topics * sizeof(double));
        part_grads_top_indep = -1 * tau * digamma(_alpha.sum);
    }
    //Update topic dependent & topic independent partial gradients
    //with contributions from this document
    ltdInd = local_topic_index;
    for (int i = 0; i < non_zero_topics; i++) {
        topic_t topic = *ltdInd;
        part_grads[topic] += digamma(_alpha.values[topic]) - digamma(
                _alpha.values[topic] + local_topic_counts[topic]);
        ++ltdInd;
    }
    part_grads_top_indep += digamma(_alpha.sum + num_words_in_doc);

    /************* Optimize **************/
    delete[] local_topic_counts;
    delete[] local_topic_index;

    return upd;
}

void* MedLDA_Model_Trainer::eval(void* token, double& eval_value) {
    return token;
}

void MedLDA_Model_Trainer::write(void* token) {
    MedLDA_update_t* upd = (MedLDA_update_t*) token;
    LDA::MedLDA_document& doc = *(upd->doc);
    doc.clear_body();
    doc.clear_gnd_label();
    LOG_IF(FATAL, !_tdoc_writer->write(doc) )<< "Couldn't write to the stream. Quitting..." << endl;
    delete upd;
}

void MedLDA_Model_Trainer::iteration_done() {
    Context& context = Context::get_instance();
    string input_w = context.get_string("input_w");
    string input_t = context.get_string("input_t");
    string output_t = context.get_string("output_t");

    //debug
    //if (taskid == 0) {
    //    for (int y = 0; y < _num_label; ++y)
    //        _eta[y].print();
    //}

    string cmd = "mv " + output_t + " " + input_t;
    system(cmd.c_str());
    //LOG(WARNING) << "Restarting IO";
    release_io();
    set_up_io(input_w, input_t, output_t);
}

void MedLDA_Model_Trainer::print_result()
{
    for (int y = 0; y < _num_label; ++y)
        _eta[y].print();
    LOG(WARNING) << "rank " << taskid << ". Accuracy: " << (double)num_accurate/_num_docs 
        << " (" << num_accurate << "/" << _num_docs << ")";
}

void* MedLDA_Model_Trainer::test(void* token) {
    LOG(FATAL) << "Test method called for trainer";
}

long MedLDA_Model_Trainer::doc_index = -1;
}
