/*******************************************************************************
    Copyright (c) 2011 Yahoo! Inc. All rights reserved.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License. See accompanying LICENSE file.

    The Initial Developer of the Original Code is Shravan Narayanamurthy.
******************************************************************************/
/*
 * MedLDA_Model_Builder.cpp
 *
 *  Created on: 06-Jan-2011
 *      
 */
#include <mpi.h>

#include "MedLDA_Model_Training_Builder.h"
#include "Context.h"
#include "DocumentReader.h"
#include "DocumentWriter.h"
#include "MedLDA_Model_Trainer.h"
#include "TopicLearner/TBB_Pipeline.h"
#include "TopicLearner/Training_Execution_Strategy.h"
#include "Local_Checkpointer.h"

#include <time.h>

namespace medlda {


MedLDA_Model_Training_Builder::MedLDA_Model_Training_Builder() {
    _dict = NULL;

    int world_rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    stringstream ss;
    ss << "." << world_rank;
    string dotrank = ss.str();

    Context& context = Context::get_instance();
    string input_prefix = context.get_string("inputprefix");
    string dict_dump = input_prefix + ".dict" + ".dump" + dotrank;
    context.put_string("dictdump", dict_dump);
    WordIndexDictionary& dict = get_dict();

    int num_words = dict.get_num_words();
    int num_topics = context.get_int("topics");

    //Params
    LOG(WARNING) 
        << "Alpha:" << context.get_double("alpha") 
        << " C:" << context.get_double("initial_c") 
        << " Ell:" << context.get_double("delta_ell") 
        << " K:" << context.get_int("topics") 
        << " Iter:" << context.get_int("iter")
        << " Sampling lag:" << context.get_int("sampling_lag");
    _model = new MedLDA_Model(num_words, num_topics);
}

MedLDA_Model_Training_Builder::~MedLDA_Model_Training_Builder() {
    delete _dict;
    delete _refiner;
    delete _pipeline;
    delete _strategy;
    delete _checkpointer;
    delete _model;
}

void MedLDA_Model_Training_Builder::init_dict() {
    _dict = new WordIndexDictionary;
    string dict_dump = Context::get_instance().get_string("dictdump");
    LOG(WARNING) << "Initializing Dictionary from " << dict_dump;
    _dict->initialize_from_dump(dict_dump);
    LOG(WARNING) << "Dictionary Initialized";
}

WordIndexDictionary& MedLDA_Model_Training_Builder::get_dict() {
    if (!_dict)
        init_dict();
    return *_dict;
}

size_t MedLDA_Model_Training_Builder::initialize_topics(string input_w,
        string input_t, int num_topics) {
    DocumentReader rdr(input_w);
    DocumentWriter wrt(input_t);
    LDA::MedLDA_document wdoc, tdoc;
    size_t num_docs = 0;
    srand(time(NULL));
    Context& context = Context::get_instance();
    int num_label = context.get_int("num_label");
    while (rdr.read(&wdoc) != -1) {
        topic_t top;
        tdoc.Clear();
        for (int i = 0; i < wdoc.body_size(); i++) {
            top = rand() % num_topics;
            tdoc.add_topic_assignment(top);
        }
        for (int y = 0; y < num_label; ++y) {
            tdoc.add_lambda(1);
        }
        wrt.write(tdoc);
        ++num_docs;
    }
    return num_docs;
}

Model_Refiner& MedLDA_Model_Training_Builder::create_model_refiner() {
    int world_rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    stringstream ss;
    ss << "." << world_rank;
    string dotrank = ss.str();

    Context& context = Context::get_instance();
    string input_prefix = context.get_string("inputprefix");

    string input_w = input_prefix + ".wor" + dotrank;
    string input_t = input_prefix + ".top" + dotrank;
    string output_t = input_prefix + ".top.tmp" + dotrank;
    int num_topics = context.get_int("topics");
    context.put_string("ignore_old_topic","false");
    size_t num_docs = initialize_topics(input_w, input_t, num_topics);
    context.put_string("input_w", input_w);
    context.put_string("input_t", input_t);
    context.put_string("output_t", output_t);

    TypeTopicCounts& ttc = _model->get_ttc();
    Parameter& alpha = _model->get_parameter(MedLDA_Model::ALPHA);
    Parameter& beta = _model->get_parameter(MedLDA_Model::BETA);
    Parameter* eta = _model->get_eta();

    _refiner = new MedLDA_Model_Trainer(ttc, alpha, beta, eta, num_docs);
    return *_refiner;
}

Pipeline& MedLDA_Model_Training_Builder::create_pipeline(
        Model_Refiner& refiner) {
    _pipeline = new TBB_Pipeline(refiner);
    return *_pipeline;
}

Execution_Strategy& MedLDA_Model_Training_Builder::create_execution_strategy(
        Pipeline& pipeline) {
    _checkpointer = new Local_Checkpointer();
    _strategy = new Training_Execution_Strategy(pipeline, *_model,
            *_checkpointer);
    return *_strategy;
}

void MedLDA_Model_Training_Builder::create_output() {
    int world_rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    stringstream ss;
    ss << "." << world_rank;
    string dotrank = ss.str();

    Context& context = Context::get_instance();
    string output_t = context.get_string("output_t");
    string cmd = "rm " + output_t;
    system(cmd.c_str());

    string input_prefix = context.get_string("inputprefix");
    string ttcdump = input_prefix + ".ttc" + ".dump" + dotrank;
    //string alphadump = input_prefix + ".alpha" + ".dump";
    string etadump = input_prefix + ".eta" + ".dump" + dotrank;
    //string etatxt = input_prefix + ".eta" + ".txt";
    context.put_string("ttcdump", ttcdump);
    //context.put_string("alphadump", alphadump);
    context.put_string("etadump", etadump);
    //context.put_string("etatxt", etatxt);
    _model->save();

    //string tw_out = input_prefix + ".topToWor" + ".txt" + dot_local;
    //context.put_string("tw_out", tw_out);
    //_model->write_statistics(get_dict());

    /*
    string tp_out = input_prefix + ".docToTop" + ".txt";
    std::ofstream tp_output(tp_out.c_str());
    LOG_IF(FATAL,!tp_output.is_open())<< "Unable to open output file: " << tp_out;
    string txt_out = input_prefix + ".worToTop" + ".txt";
    std::ofstream txt_output(txt_out.c_str());
    LOG_IF(FATAL,!txt_output.is_open()) << "Unable to open output file: " << txt_out;
    LOG(WARNING) << "Saving document to topic-proportions in " << tp_out;
    LOG(WARNING) << "Saving word to topic assignment in " << txt_out;

    string input_w = context.get_string("input_w");
    string input_t = context.get_string("input_t");
    int num_topics = context.get_int("topics");
    DocumentReader wdoc_rdr(input_w), tdoc_rdr(input_t);
    LDA::MedLDA_document doc;
    while(wdoc_rdr.read(&doc)!=-1) {
        LDA::MedLDA_document tdoc;
        tdoc_rdr.read(&tdoc);
        doc.MergeFrom(tdoc);

        //For each document
        int num_words_in_doc = doc.body_size();
        {
            //Write Topic Proportions
            vector<tppair> topProb;
            //Write first entry
            tp_output << doc.docid() << "\t" << doc.binary_label() << "\t";

            topic_t* local_topic_counts = (topic_t*)calloc(num_topics,sizeof(topic_t));

            for(int k=0;k<num_words_in_doc;k++) {
                topic_t topic = doc.topic_assignment(k);
                ++local_topic_counts[topic];
            }

            //Store the topic, counts as proportion in a vector
            topic_t* ltcInd = local_topic_counts;
            for(topic_t t=0;t<num_topics;t++) {
                int n = *ltcInd;
                if(n!=0) {
                    tppair tp(t,(float)n/num_words_in_doc);
                    topProb.push_back(tp);
                }
                ++ltcInd;
            }
            free(local_topic_counts);

            //Sort on the counts proportion
            std::sort(topProb.begin(),topProb.end(),prob_cmp);

            //Write out the second entry
            for(int i=0;i<min((int)topProb.size(),NUM_TOPICS_PER_DOC);i++) {
                tp_output << "(" << topProb[i].first << "," << topProb[i].second << ")";
                if(i<min((int)topProb.size(),NUM_TOPICS_PER_DOC)-1)
                tp_output << " ";
            }
            tp_output << endl;
        }
        {
            //Write word to topic assignments
            txt_output << doc.docid() << "\t" << doc.binary_label() << "\t";
            for(int i=0;i<num_words_in_doc;i++) {
                txt_output << "(" << _dict->get_word(doc.body(i)) << "," << doc.topic_assignment(i) << ")";
                if(i<num_words_in_doc-1)
                txt_output << " ";
            }
            txt_output << endl;

        }
    }
    tp_output.flush();
    tp_output.close();
    txt_output.flush();
    txt_output.close();

    LOG(WARNING) << "Document to topic-proportions saved in " << tp_out;
    LOG(WARNING) << "Word to topic assignment saved in " << txt_out;
    */
}

void MedLDA_Model_Training_Builder::print_result() {
    Context& context = Context::get_instance();
    string input_w = context.get_string("input_w");
    string input_t = context.get_string("input_t");
    int _num_label = context.get_int("num_label");
    int _num_topics = context.get_int("topics");

    Parameter& alpha = _model->get_parameter(MedLDA_Model::ALPHA);
    Parameter* eta = _model->get_eta();
    DocumentReader wr(input_w);
    DocumentReader tr(input_t);
    LDA::MedLDA_document doc, tdoc;

    topic_t *document_topic_counts = new topic_t[_num_topics];
    int num_accurate = 0;
    int num_docs = 0;
    int stat[_num_label];
    memset(stat, 0, sizeof(int) * _num_label);
    while (wr.read(&doc) != -1) {
        tr.read(&tdoc);
        doc.MergeFrom(tdoc);
        //Document statistics
        int num_words_in_doc = doc.body_size();
        memset(document_topic_counts, 0, sizeof(topic_t)*_num_topics);
        for (int k = 0; k < num_words_in_doc; k++) {
            topic_t topic = doc.topic_assignment(k);
            ++document_topic_counts[topic];
        }
        double score, max_score = 0;
        double *alpha_array = alpha.values;
        double *eta_array;
        int pred = -1;
        for (int y = 0; y < _num_label; ++y) {
            eta_array = eta[y].values;
            score = 0;
            for (topic_t t = 0; t < _num_topics; ++t) 
                score += eta_array[t] * (double)(document_topic_counts[t] + alpha_array[t])
                        / (num_words_in_doc + alpha.sum);
            if (pred == -1 || score > max_score) {
                pred = y;
                max_score = score;
            }
        }
        bool in = false;
        for (int i = 0; i < doc.gnd_label_size(); ++i) {
            if (doc.gnd_label(i) == pred) {
                in = true;
            }
        }
        if (in)
            ++num_accurate;

        ++stat[pred];
        ++num_docs;
    }
    delete[] document_topic_counts;
    
    //MPI
    char processor_name[MPI_MAX_PROCESSOR_NAME];
    int name_len;
    MPI_Get_processor_name(processor_name, &name_len);
    int world_rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    stringstream ss;
    ss << "At " << processor_name << " (rank " << world_rank << "). ";
    string nodeinfo = ss.str();
    LOG(WARNING) << nodeinfo << "Top 1 in gnd list: " 
        << (double)num_accurate/num_docs 
        << " (" << num_accurate << "/" << num_docs << ")";

    ss.str("");
    for (int y = 0; y < _num_label; ++y) {
        ss << y << ":" << stat[y] << " ";
    }
    LOG(WARNING) << nodeinfo << "stat: " << ss.str();
}

Model& MedLDA_Model_Training_Builder::get_model() {
    return *_model;
}

}
